import enum

AUDIO_FORMAT_PARAMETER = 'format'

class ParameterType(enum.Enum):
    number = enum.auto() # floating point number
    number_int = enum.auto() # integer number
    list = enum.auto() # list of possible string values

class AudioFormat(enum.Enum):
    mp3 = enum.auto()
    ogg_opus = enum.auto()
    ogg_vorbis = enum.auto()


import pydantic
import logging
import pprint
import pydub
import tempfile
from pydantic import Field
from typing import Optional
import cloudlanguagetools.servicemanager
import cloudlanguagetools.options
import cloudlanguagetools.languages

logger = logging.getLogger(__name__)

"""
This API is meant to interface with chatbots such as ChatGPT. 
it has a simplified interface meant to integrate well with OpenAI functions, and returns text only
(with the exception of audio)
it will try to find the ideal service and parameters for the given request, 
but sometimes parameters will be overriden to ensure an output is produced.
"""

class NoDataFoundException(Exception):
    pass

class TranslateLookupQuery(pydantic.BaseModel):
    input_text: str = Field(description="text to translate or lookup in the dictionary")
    source_language: cloudlanguagetools.languages.CommonLanguage = Field(description="language of the text to translate or lookup in dictionary")
    target_language: cloudlanguagetools.languages.CommonLanguage = Field(description="language to translate to, or language of dictionary definition")
    service: Optional[cloudlanguagetools.constants.Service] = Field(default=None, description='service to use for translation')

class TransliterateQuery(pydantic.BaseModel):
    input_text: str = Field(description="text to transliterate")
    language: cloudlanguagetools.languages.CommonLanguage = Field(description="language the text is in")
    service: Optional[cloudlanguagetools.constants.Service] = Field(default=None, description='service to use for transliteration')

class AudioQuery(pydantic.BaseModel):
    input_text: str = Field(description="text to pronounce / generate audio")
    language: cloudlanguagetools.languages.CommonLanguage = Field(description="language the text is in")
    service: Optional[cloudlanguagetools.constants.Service] = Field(default=None, description='service to use audio pronunciation')
    gender: Optional[cloudlanguagetools.constants.Gender] = Field(default=None, description='gender of the voice to pronounce the text in')

class BreakdownQuery(pydantic.BaseModel):
    input_text: str = Field(description="text to breakdown into words")
    language: cloudlanguagetools.languages.CommonLanguage = Field(description="language the text is in")
    translation_language: cloudlanguagetools.languages.CommonLanguage = Field(default=cloudlanguagetools.languages.CommonLanguage.en, description="language to translate into")
    translation_service: Optional[cloudlanguagetools.constants.Service] = Field(default=None, description='service to use for translation')
    transliteration_service: Optional[cloudlanguagetools.constants.Service] = Field(default=None, description='service to use for transliteration')


class ChatAPI():
    def __init__(self, manager):
        self.manager = manager

    def get_service_preference(self, preferred_service_list, default_service):
        if default_service != None:
            return [default_service] + preferred_service_list
        else:
            return preferred_service_list

    def select_translation_option(self, preferred_service: cloudlanguagetools.constants.Service,
            source_language: cloudlanguagetools.languages.Language,
            target_language: cloudlanguagetools.languages.Language):   
        """pick appropriate translation service"""

        service_preference = self.get_service_preference([
            cloudlanguagetools.constants.Service.DeepL,
            cloudlanguagetools.constants.Service.Azure,
            cloudlanguagetools.constants.Service.Google,
            cloudlanguagetools.constants.Service.Amazon,
            cloudlanguagetools.constants.Service.Watson            
        ], preferred_service)

        # get list of translation options
        translation_language_list = self.manager.get_translation_language_list()
        source_translation_language_list = [x for x in translation_language_list if x.language == source_language]
        target_translation_language_list = [x for x in translation_language_list if x.language == target_language]
        # get the list of services in common between source_translation_language_list and target_translation_language_list
        source_service_list = set([x.service for x in source_translation_language_list])
        target_service_list = set([x.service for x in target_translation_language_list])
        common_service_list = source_service_list.intersection(target_service_list)

        while service_preference[0] not in common_service_list:
            service_preference.pop(0)
            if len(service_preference) == 0:
                raise NoDataFoundException(f'No service found for translation from {source_language} to {target_language}')

        service = service_preference[0]
        source_language_id = [x for x in source_translation_language_list if x.service == service][0].get_language_id()
        target_language_id = [x for x in target_translation_language_list if x.service == service][0].get_language_id()

        translation_option = {
            'service': service,
            'source_language_id': source_language_id,
            'target_language_id': target_language_id
        }
        return translation_option

    def select_transliteration_option(self, preferred_service: cloudlanguagetools.constants.Service,
            language: cloudlanguagetools.languages.Language):
        transliteration_option_list = self.manager.get_transliteration_language_list()
        candidates = [x for x in transliteration_option_list if x.language == language]
        if len(candidates) == 0:
            raise NoDataFoundException(f'No transliteration service found for language {language.lang_name}')

        service_list = set([x.service for x in candidates])

        service_preference = self.get_service_preference([
            cloudlanguagetools.constants.Service.MandarinCantonese, # in case input text is chinese
            cloudlanguagetools.constants.Service.EasyPronunciation,
            cloudlanguagetools.constants.Service.Azure,
            cloudlanguagetools.constants.Service.PyThaiNLP,
        ], preferred_service)

        while service_preference[0] not in service_list:
            service_preference.pop(0)
            if len(service_preference) == 0:
                raise NoDataFoundException(f'No service found for transliteration of {language.lang_name}')
            
        service = service_preference[0]
        final_candidates = [x for x in candidates if x.service == service]

        if service == cloudlanguagetools.constants.Service.MandarinCantonese:
            final_candidates = [x for x in candidates if 
                                x.service == service and x.get_transliteration_key()['tone_numbers'] == False and
                                x.service == service and x.get_transliteration_key()['spaces'] == False]
        transliteration_option = final_candidates[0]

        return transliteration_option

    def translate(self, query: TranslateLookupQuery):
        logger.info(f'translating {query.input_text} from {query.source_language} to {query.target_language}')

        source_language = cloudlanguagetools.languages.Language[query.source_language.name]
        target_language = cloudlanguagetools.languages.Language[query.target_language.name]
        translation_option = self.select_translation_option(query.service, source_language, target_language)

        # get the translation
        translated_text = self.manager.get_translation(
            query.input_text,
            translation_option['service'],
            translation_option['source_language_id'],
            translation_option['target_language_id'],
        )
        return translated_text


    def transliterate(self, query: TransliterateQuery):
        language = cloudlanguagetools.languages.Language[query.language.name]
        transliteration_option = self.select_transliteration_option(query.service, language)
        transliterated_text = self.manager.get_transliteration(
            query.input_text,
            transliteration_option.service,
            transliteration_option.get_transliteration_key()
        )
        return transliterated_text

    def translate_or_lookup(self, query: TranslateLookupQuery):
        logger.info(f'translation or dictionary lookup: {query}')

        try:
            # try dictionary lookup as a first attempt
            result = self.dictionary_lookup(query)
            return result
        except cloudlanguagetools.errors.NotFoundError as e:
            # if dictionary lookup fails, try translation
            result = self.translate(query)
            return result

    def dictionary_lookup(self, query: TranslateLookupQuery):
        logger.info(f'dictionary lookup {query}')
        source_language = cloudlanguagetools.languages.Language[query.source_language.name]
        target_language = cloudlanguagetools.languages.Language[query.target_language.name]
        dictionary_option_list = self.manager.get_dictionary_lookup_options()
        candidates = [x for x in dictionary_option_list if x.language == source_language and x.target_language == target_language]
        if len(candidates) == 0:
            raise NoDataFoundException(f'No dictionary service found for source language {query.source_language.lang_name} / target language: {query.target_language.lang_name}')

        service_list = set([x.service for x in candidates])

        preferred_service = query.service
        # for Chinese, always enforce Wenlin
        if source_language in [
            cloudlanguagetools.languages.Language.yue,
            cloudlanguagetools.languages.Language.zh_cn,
            cloudlanguagetools.languages.Language.zh_tw,
            cloudlanguagetools.languages.Language.zh_lit
        ]:
            preferred_service = cloudlanguagetools.constants.Service.Wenlin

        service_preference = self.get_service_preference([
            cloudlanguagetools.constants.Service.Wenlin,
            cloudlanguagetools.constants.Service.Azure,
        ], preferred_service)

        while service_preference[0] not in service_list:
            service_preference.pop(0)
            if len(service_preference) == 0:
                raise NoDataFoundException(f'No service found for dictionary lookup of {query.source_language.lang_name}')
            
        service = service_preference[0]
        final_candidates = [x for x in candidates if x.service == service]

        dictionary_option = final_candidates[0]
        logger.debug(f'Using dictionary option {pprint.pformat(dictionary_option.json_obj())}')

        dictionary_result = self.manager.get_dictionary_lookup(
            query.input_text,
            service,
            dictionary_option.get_lookup_key()
        )
        result = ' / '.join(dictionary_result)
        return result
        

    def audio(self, query: AudioQuery, format: cloudlanguagetools.options.AudioFormat) -> tempfile.NamedTemporaryFile:
        logger.info(f'processing audio query: {query}')
        language = cloudlanguagetools.languages.Language[query.language.name]
        # get full voice list, filter down to correct language
        # ====================================================
        voice_list = self.manager.get_tts_voice_list()
        default_audio_language = cloudlanguagetools.languages.AudioLanguageDefaults[language]
        candidates = [x for x in voice_list if x.audio_language == default_audio_language]

        # select service
        # ==============

        service_list = set([x.service for x in candidates])
        service_preference = self.get_service_preference([
            cloudlanguagetools.constants.Service.Azure,
            cloudlanguagetools.constants.Service.Amazon,
            cloudlanguagetools.constants.Service.Google,
            cloudlanguagetools.constants.Service.Watson,
            cloudlanguagetools.constants.Service.Naver,
            cloudlanguagetools.constants.Service.CereProc,
            cloudlanguagetools.constants.Service.VocalWare,
            cloudlanguagetools.constants.Service.FptAi,
        ], query.service)

        while service_preference[0] not in service_list:
            service_preference.pop(0)
            if len(service_preference) == 0:
                raise NoDataFoundException(f'No service found for audio pronouncation of {language.lang_name}')

        # restrict to candidates for that service
        service = service_preference[0]
        candidates = [x for x in candidates if x.service == service]

        # select gender
        # =============
        gender_list = set([x.gender for x in candidates])
        gender_preference = [
            cloudlanguagetools.constants.Gender.Female,
            cloudlanguagetools.constants.Gender.Male,
            cloudlanguagetools.constants.Gender.Any
        ]
        if query.gender != None:
            gender_preference = [query.gender] + gender_preference

        while gender_preference[0] not in gender_list:
            gender_preference.pop(0)
        gender = gender_preference[0]
        logger.debug(f'selected gender: {gender}')

        candidates = [x for x in candidates if x.gender == gender]

        # pick the first candidate and generate audio
        # ===========================================

        voice = candidates[0]
        logger.debug(f'picked voice: {voice.get_voice_description()}')

        options = {}
        convert_mp3_to_ogg = False
        # by default, the format is mp3. if the request format is ogg, then we need to convert
        if format == cloudlanguagetools.options.AudioFormat.ogg_opus:
            logger.debug(f'requested ogg_opus format')
            convert_mp3_to_ogg = True
            if cloudlanguagetools.options.AUDIO_FORMAT_PARAMETER in voice.get_options():
                logger.debug(f'voice supports audio format parameter')
                voice_available_formats = voice.get_options()[cloudlanguagetools.options.AUDIO_FORMAT_PARAMETER]['values']
                logger.debug(f'voice available formats: {voice_available_formats}')
                if format.name in voice_available_formats:
                    # native ogg opus supported
                    logger.debug(f'ogg opus natively supported')
                    options[cloudlanguagetools.options.AUDIO_FORMAT_PARAMETER] = format.name
                    convert_mp3_to_ogg = False

        # generate audio
        logger.debug(f'generating audio with voice {pprint.pformat(voice.json_obj())} options {pprint.pformat(options)}')
        audio_temp_file = self.manager.get_tts_audio(
            query.input_text,
            service,
            voice.get_voice_key(),
            options
        )

        if convert_mp3_to_ogg:
            logger.debug(f'need to convert from mp3 to ogg_opus')
            audio = pydub.AudioSegment.from_mp3(audio_temp_file.name)
            ogg_audio_temp_file = tempfile.NamedTemporaryFile(prefix='cloudlanguagetools_chatapi', suffix='.ogg')
            audio.export(ogg_audio_temp_file.name, format="ogg", codec="libopus")
            audio_temp_file = ogg_audio_temp_file

        return audio_temp_file

    def recognize_audio(self, sound_temp_file: tempfile.NamedTemporaryFile, audio_format: cloudlanguagetools.options.AudioFormat):
        # logger.debug(f'processing audio query: {query}')
        # result = self.manager.services[cloudlanguagetools.constants.Service.Azure].speech_to_text(sound_temp_file.name, audio_format)
        result = self.manager.services[cloudlanguagetools.constants.Service.OpenAI].speech_to_text(sound_temp_file.name, audio_format)
        return result

    def breakdown(self, query: BreakdownQuery):
        logger.debug(f'processing breakdown query: {query}')

        language = cloudlanguagetools.languages.Language[query.language.name]
        translation_language = cloudlanguagetools.languages.Language[query.translation_language.name]

        # locate tokenization option
        # ==========================
        tokenization_options = self.manager.get_tokenization_options()
        tokenization_candidates = [x for x in tokenization_options if x.language == language]
        if len(tokenization_candidates) == 0:
            raise NoDataFoundException(f'No tokenization options found for language {language.lang_name}')
        tokenization_option = tokenization_candidates[0]

        # locate translation option
        # =========================
        translation_option = self.select_translation_option(query.translation_service, language, translation_language)

        # locate transliteration option
        # =============================
        transliteration_option = self.select_transliteration_option(query.transliteration_service, language)

        breakdown_result = self.manager.get_breakdown(query.input_text, 
            tokenization_option.json_obj(), 
            translation_option, 
            transliteration_option.json_obj())

        # process breakdown result
        # ========================

        lines = []
        for entry in breakdown_result:
            line_components = []
            line_components.append(entry['token'] + ':')
            if 'lemma' in entry and entry['lemma'] != entry['token']:
                line_components.append('/' + entry['lemma'])
            
            if 'transliteration' in entry:
                line_components.append(entry['transliteration'] + ',')

            if 'translation' in entry:
                line_components.append(entry['translation'])

            if 'pos_description' in entry:
                line_components.append('(' + entry['pos_description'] + ')')
            
            lines.append(' '.join(line_components))

        return '\n'.join(lines)



class TranslationLanguage():
    def __init__(self):
        pass

    def get_language_code(self):
        return self.language.name

    def get_language_name(self):
        return self.language.lang_name

    def json_obj(self):
        return {
            'service': self.service.name,
            'language_code': self.get_language_code(),
            'language_name': self.get_language_name(),
            'language_id': self.get_language_id(),
            'service_fee': self.service_fee.name
        }    


import json
import requests
import tempfile
import logging
import pprint

import cloudlanguagetools.service
import cloudlanguagetools.constants
import cloudlanguagetools.languages
import cloudlanguagetools.ttsvoice
import cloudlanguagetools.translationlanguage
import cloudlanguagetools.transliterationlanguage
import cloudlanguagetools.errors

logger = logging.getLogger(__name__)

def get_translation_language_enum(language_id):
    # print(f'language_id: {language_id}')
    watson_language_id_map = {
        'fr-CA': 'fr_ca',
        'id': 'id_',
        'pt': 'pt_pt',
        'sr': 'sr_cyrl',
        'zh':'zh_cn',
        'zh-TW': 'zh_tw'

    }
    if language_id in watson_language_id_map:
        language_id = watson_language_id_map[language_id]
    return cloudlanguagetools.languages.Language[language_id]

def get_audio_language_enum(voice_language):
    watson_audio_id_map = {
        'ar-MS': 'ar_XA'
    }
    language_enum_name = voice_language.replace('-', '_')
    if voice_language in watson_audio_id_map:
        language_enum_name = watson_audio_id_map[voice_language]
    return cloudlanguagetools.languages.AudioLanguage[language_enum_name]

class WatsonTranslationLanguage(cloudlanguagetools.translationlanguage.TranslationLanguage):
    def __init__(self, language_id):
        self.service = cloudlanguagetools.constants.Service.Watson
        self.service_fee = cloudlanguagetools.constants.ServiceFee.paid
        self.language_id = language_id
        self.language = get_translation_language_enum(language_id)

    def get_language_id(self):
        return self.language_id

class WatsonVoice(cloudlanguagetools.ttsvoice.TtsVoice):
    def __init__(self, voice_data):
        self.service = cloudlanguagetools.constants.Service.Watson
        self.service_fee = cloudlanguagetools.constants.ServiceFee.paid
        self.audio_language = get_audio_language_enum(voice_data['language'])
        self.name = voice_data['name']
        self.description = voice_data['description']
        self.description = voice_data['description']
        self.gender = cloudlanguagetools.constants.Gender[voice_data['gender'].capitalize()]


    def get_voice_key(self):
        return {
            'name': self.name
        }

    def get_voice_shortname(self):
        is_dnn = ''
        if 'Dnn' in self.description:
            is_dnn = ' (Dnn)'
        return self.description.split(':')[0] + is_dnn

    def get_options(self):
        return {}

class WatsonService(cloudlanguagetools.service.Service):
    def __init__(self):
        pass

    def configure(self, config):
        self.translator_key = config['translator_api_key']
        self.translator_url = config['translator_url']
        self.speech_key = config['speech_api_key']
        self.speech_url = config['speech_url']
    
    def get_tts_voice_list(self):
        return []

    def get_translation_languages(self):
        response = requests.get(self.translator_url + '/v3/languages?version=2018-05-01', auth=('apikey', self.translator_key), timeout=cloudlanguagetools.constants.RequestTimeout)
        return response.json()

    def get_translation_language_list(self):
        language_list = self.get_translation_languages()['languages']
        result = []
        # print(language_list)
        for entry in language_list:
            if entry['supported_as_source'] == True and entry['supported_as_target'] == True:
                # print(entry)
                language_id = entry['language']
                try:
                    result.append(WatsonTranslationLanguage(language_id))
                except KeyError:
                    logging.error(f'could not process translation language for {language_id}, {entry}', exc_info=True)                    
        return result        

    def list_voices(self):
        response = requests.get(self.speech_url + '/v1/voices', auth=('apikey', self.speech_key), timeout=cloudlanguagetools.constants.RequestTimeout)
        data = response.json()
        logger.debug(f'voices: {data}')
        return data['voices']

    def get_tts_voice_list(self):
        result = []

        voice_list = self.list_voices()
        for voice in voice_list:
            try:
                result.append(WatsonVoice(voice))
                logger.debug(f'added voice {voice}')
            except KeyError:
                logging.error(f'could not process voice for {voice}', exc_info=True)

        return result

    def get_tts_audio(self, text, voice_key, options):
        output_temp_file = tempfile.NamedTemporaryFile()
        output_temp_filename = output_temp_file.name

        base_url = self.speech_url
        url_path = '/v1/synthesize'
        voice_name = voice_key["name"]
        constructed_url = base_url + url_path + f'?voice={voice_name}'
        headers = {
            'Content-Type': 'application/json',
            'Accept': 'audio/mp3'
        }

        data = {
            'text': text
        }

        response = requests.post(constructed_url, data=json.dumps(data), auth=('apikey', self.speech_key), headers=headers, timeout=cloudlanguagetools.constants.RequestTimeout)

        if response.status_code == 200:
            with open(output_temp_filename, 'wb') as audio:
                audio.write(response.content)
            return output_temp_file

        # otherwise, an error occured
        error_message = f"Status code: {response.status_code} reason: {response.reason} voice: [{voice_name}]]"
        raise cloudlanguagetools.errors.RequestError(error_message)


    def get_transliteration_language_list(self):
        return []

    def get_translation(self, text, from_language_key, to_language_key):
        body = {
            'text': text,
            'source': from_language_key,
            'target': to_language_key
        }
        response = requests.post(self.translator_url + '/v3/translate?version=2018-05-01', auth=('apikey', self.translator_key), json=body, timeout=cloudlanguagetools.constants.RequestTimeout)

        if response.status_code == 200:
            # {'translations': [{'translation': 'Le coût est très bas.'}], 'word_count': 2, 'character_count': 4}
            data = response.json()
            return data['translations'][0]['translation']

        error_message = error_message = f'Watson: could not translate text [{text}] from {from_language_key} to {to_language_key} ({response.json()})'
        raise cloudlanguagetools.errors.RequestError(error_message)


import json
import pprint
import requests
import tempfile
import os
import contextlib
import logging
from typing import List

import cloudlanguagetools.service
import cloudlanguagetools.constants
import cloudlanguagetools.options
import cloudlanguagetools.languages
import cloudlanguagetools.ttsvoice
import cloudlanguagetools.translationlanguage
import cloudlanguagetools.transliterationlanguage
import cloudlanguagetools.errors

logger = logging.getLogger(__name__)


DEFAULT_STABILITY = 0.75
DEFAULT_SIMILARITY_BOOST = 0.75

GENDER_MAP = {
    'male': cloudlanguagetools.constants.Gender.Male,
    'female': cloudlanguagetools.constants.Gender.Female,
}

VOICE_OPTIONS = {
            'stability' : {
                'type': cloudlanguagetools.options.ParameterType.number.name,
                'min': 0.0,
                'max': 1.0,
                'default': DEFAULT_STABILITY
            },
            'similarity_boost' : {
                'type': cloudlanguagetools.options.ParameterType.number.name,
                'min': 0.0,
                'max': 1.0,
                'default': DEFAULT_SIMILARITY_BOOST
            },                        
}

class ElevenLabsVoice(cloudlanguagetools.ttsvoice.TtsVoice):
    def __init__(self, voice_data, language: cloudlanguagetools.languages.AudioLanguage, model_id, model_short_name):
        # pprint.pprint(voice_data)
        self.service = cloudlanguagetools.constants.Service.ElevenLabs
        self.service_fee = cloudlanguagetools.constants.ServiceFee.paid
        self.voice_id = voice_data['voice_id']
        self.model_id = model_id
        self.model_short_name = model_short_name
        self.name = voice_data['name']
        self.gender = GENDER_MAP.get(voice_data['labels']['gender'], cloudlanguagetools.constants.Gender.Male)
        self.audio_language = language

    def get_voice_key(self):
        return {
            'voice_id': self.voice_id,
            'model_id': self.model_id,
            'language': self.audio_language.name
        }

    def get_voice_shortname(self):
        return f'{self.name} ({self.model_short_name})'

    def get_options(self):
        return VOICE_OPTIONS

class ElevenLabsService(cloudlanguagetools.service.Service):
    def __init__(self):
        self.service = cloudlanguagetools.constants.Service.ElevenLabs

    def configure(self, config):
        self.api_key = config['api_key']

    def get_headers(self):
        return {
            "Accept": "application/json",
            "xi-api-key": self.api_key
        }

    def get_tts_audio(self, text, voice_key, options):
        voice_id = voice_key['voice_id']
        url = f'https://api.elevenlabs.io/v1/text-to-speech/{voice_id}'

        headers = self.get_headers()
        headers['Accept'] = "audio/mpeg"

        data = {
            "text": text,
            "model_id": voice_key['model_id'],
            "voice_settings": {
                "stability": options.get('stability', DEFAULT_STABILITY),
                "similarity_boost": options.get('similarity_boost', DEFAULT_SIMILARITY_BOOST)
            }
        }

        return self.get_tts_audio_base_post_request(url, json=data, headers=headers)



    def get_audio_language(self, language_id) -> cloudlanguagetools.languages.AudioLanguage:
        logger.debug(f'processing language_id: {language_id}')
        override_map = {
            'pt': cloudlanguagetools.languages.AudioLanguage.pt_PT,
            'en-uk': cloudlanguagetools.languages.AudioLanguage.en_GB,
            'zh': cloudlanguagetools.languages.AudioLanguage.zh_CN,
            'id': cloudlanguagetools.languages.AudioLanguage.id_ID
        }
        if language_id in override_map:
            return override_map[language_id]
        # try to reconstruct AudioLanguage
        language_id_components = language_id.split('-')
        if len(language_id_components) != 2:
            language_enum = cloudlanguagetools.languages.Language[language_id]
            audio_language_enum = cloudlanguagetools.languages.AudioLanguageDefaults[language_enum]
            return audio_language_enum
        else:
            # language is in format en-us
            modified_language_id = language_id_components[0] + '_' + language_id_components[1].upper()
            logger.debug(f'modified_language_id: {modified_language_id}')
            audio_language_enum = cloudlanguagetools.languages.AudioLanguage[modified_language_id]
            return audio_language_enum

    def get_tts_voice_list(self) -> List[ElevenLabsVoice]:
        result = []

        # first, get all models to get list of languages
        url = "https://api.elevenlabs.io/v1/models"
        response = requests.get(url, headers=self.get_headers(), timeout=cloudlanguagetools.constants.RequestTimeout)
        response.raise_for_status()
        model_data = response.json()

        # restrict to models that can do text to speech (elevenlabs introduced voice conversion)
        model_data = [model for model in model_data if model['can_do_text_to_speech']]

        #pprint.pprint(model_data)
        # model_data: 
        # [{'can_be_finetuned': True,
        # 'can_do_text_to_speech': True,
        # 'can_do_voice_conversion': False,
        # 'description': 'Use our standard English language model to generate speech '
        #                 'in a variety of voices, styles and moods.',
        # 'languages': [{'language_id': 'en', 'name': 'English'}],
        # 'model_id': 'eleven_monolingual_v1',
        # 'name': 'Eleven Monolingual v1',
        # 'token_cost_factor': 1.0},
        # {'can_be_finetuned': True,
        # 'can_do_text_to_speech': True,
        # 'can_do_voice_conversion': True,
        # 'description': 'Generate lifelike speech in multiple languages and create '
        #                 'content that resonates with a broader audience. ',
        # 'languages': [{'language_id': 'en', 'name': 'English'},
        #                 {'language_id': 'de', 'name': 'German'},
        #                 {'language_id': 'pl', 'name': 'Polish'},
        #                 {'language_id': 'es', 'name': 'Spanish'},
        #                 {'language_id': 'it', 'name': 'Italian'},
        #                 {'language_id': 'fr', 'name': 'French'},
        #                 {'language_id': 'pt', 'name': 'Portuguese'},
        #                 {'language_id': 'hi', 'name': 'Hindi'}],
        # 'model_id': 'eleven_multilingual_v1',
        # 'name': 'Eleven Multilingual v1',
        # 'token_cost_factor': 1.0}]
        #         


        # now, retrieve voice list
        # call elevenlabs API to list TTS voices
        url = "https://api.elevenlabs.io/v1/voices"

        response = requests.get(url, headers=self.get_headers(), timeout=cloudlanguagetools.constants.RequestTimeout)
        response.raise_for_status()

        data = response.json()

        for model in model_data:
            model_id = model['model_id']
            model_name = model['name']
            model_short_name = model_name.replace('Eleven ', '').strip()
            for language_record in model['languages']:
                try:
                    language_id = language_record['language_id']
                    audio_language_enum = self.get_audio_language(language_id)
                    for voice_data in data['voices']:
                        result.append(ElevenLabsVoice(voice_data, audio_language_enum, model_id, model_short_name))
                except Exception as e:
                    logger.exception(f'ElevenLabs: error processing voice_data: {voice_data}')

        return result


    def get_tts_voice_list_v3(self) -> List[cloudlanguagetools.ttsvoice.TtsVoice_v3]:
        result = []

        # first, get all models to get list of languages
        url = "https://api.elevenlabs.io/v1/models"
        response = requests.get(url, headers=self.get_headers(), timeout=cloudlanguagetools.constants.RequestTimeout)
        response.raise_for_status()
        model_data = response.json()

        # restrict to models that can do text to speech (elevenlabs introduced voice conversion)
        model_data = [model for model in model_data if model['can_do_text_to_speech']]

        #pprint.pprint(model_data)
        # model_data: 
        # [{'can_be_finetuned': True,
        # 'can_do_text_to_speech': True,
        # 'can_do_voice_conversion': False,
        # 'description': 'Use our standard English language model to generate speech '
        #                 'in a variety of voices, styles and moods.',
        # 'languages': [{'language_id': 'en', 'name': 'English'}],
        # 'model_id': 'eleven_monolingual_v1',
        # 'name': 'Eleven Monolingual v1',
        # 'token_cost_factor': 1.0},
        # {'can_be_finetuned': True,
        # 'can_do_text_to_speech': True,
        # 'can_do_voice_conversion': True,
        # 'description': 'Generate lifelike speech in multiple languages and create '
        #                 'content that resonates with a broader audience. ',
        # 'languages': [{'language_id': 'en', 'name': 'English'},
        #                 {'language_id': 'de', 'name': 'German'},
        #                 {'language_id': 'pl', 'name': 'Polish'},
        #                 {'language_id': 'es', 'name': 'Spanish'},
        #                 {'language_id': 'it', 'name': 'Italian'},
        #                 {'language_id': 'fr', 'name': 'French'},
        #                 {'language_id': 'pt', 'name': 'Portuguese'},
        #                 {'language_id': 'hi', 'name': 'Hindi'}],
        # 'model_id': 'eleven_multilingual_v1',
        # 'name': 'Eleven Multilingual v1',
        # 'token_cost_factor': 1.0}]
        #         


        # now, retrieve voice list
        # call elevenlabs API to list TTS voices
        url = "https://api.elevenlabs.io/v1/voices"

        response = requests.get(url, headers=self.get_headers(), timeout=cloudlanguagetools.constants.RequestTimeout)
        response.raise_for_status()

        data = response.json()

        for model in model_data:
            model_id = model['model_id']
            model_name = model['name']
            model_short_name = model_name.replace('Eleven ', '').strip()
            # for language_record in model['languages']:
            for voice_data in data['voices']:
                voice_name = voice_data['name']
                try:
                    languages = model['languages']
                    language_id_list = [language_record['language_id'] for language_record in languages]
                    audio_language_enum_list = [self.get_audio_language(language_id) for language_id in language_id_list]
                    voice = cloudlanguagetools.ttsvoice.TtsVoice_v3(
                        name=f'{voice_name} ({model_short_name})',
                        voice_key={
                            'voice_id': voice_data['voice_id'],
                            'model_id': model_id,
                        },
                        options=VOICE_OPTIONS,
                        service=cloudlanguagetools.constants.Service.ElevenLabs,
                        gender=GENDER_MAP.get(voice_data['labels']['gender'], cloudlanguagetools.constants.Gender.Male),
                        audio_languages=audio_language_enum_list,
                        service_fee=cloudlanguagetools.constants.ServiceFee.paid
                    )
                    result.append(voice)
                except Exception as e:
                    logger.exception(f'ElevenLabs: error processing voice_data: {voice_data}')

        return result




import json
import requests
import tempfile
import logging
import clt_wenlin
import sqlite3

import cloudlanguagetools.service
import cloudlanguagetools.constants
import cloudlanguagetools.languages
import cloudlanguagetools.errors
import cloudlanguagetools.dictionarylookup

class WenlinDictionaryLookup(cloudlanguagetools.dictionarylookup.DictionaryLookup):
    def __init__(self, source_language, lookup_type):
        self.service = cloudlanguagetools.constants.Service.Wenlin
        self.service_fee = cloudlanguagetools.constants.ServiceFee.free
        self.target_language = cloudlanguagetools.languages.Language.en
        self.language = source_language
        self.lookup_type = lookup_type

    def get_lookup_key(self):
        return {
            'language': self.language.name,
            'lookup_type': self.lookup_type.name
        }        

    def get_lookup_name(self):
        return f'Wenlin ({self.language.lang_name} to {self.target_language.lang_name}), {self.lookup_type.name}'

    def get_lookup_shortname(self):
        return f'Wenlin, {self.lookup_type.name}'


class WenlinService(cloudlanguagetools.service.Service):
    def __init__(self):
        pass

    def configure(self, config):
        pass
    
    def get_tts_voice_list(self):
        return []

    def get_translation_language_list(self):
        return []

    def get_tts_voice_list(self):
        return []

    def get_tts_audio(self, text, voice_key, options):
        raise Exception('not supported')

    def get_transliteration_language_list(self):
        return []

    def get_translation(self, text, from_language_key, to_language_key):
        raise Exception('not supported')

    def get_dictionary_lookup_list(self):
        result = []
        for language in [
            cloudlanguagetools.languages.Language.zh_cn,
            cloudlanguagetools.languages.Language.zh_tw,
            cloudlanguagetools.languages.Language.yue
        ]:
            result.extend([
                WenlinDictionaryLookup(language, cloudlanguagetools.constants.DictionaryLookupType.Definitions),
                WenlinDictionaryLookup(language, cloudlanguagetools.constants.DictionaryLookupType.PartOfSpeech),
                WenlinDictionaryLookup(language, cloudlanguagetools.constants.DictionaryLookupType.MeasureWord),
                WenlinDictionaryLookup(language, cloudlanguagetools.constants.DictionaryLookupType.PartOfSpeechDefinitions),
            ])

        return result

    def get_connection(self):
        db_filepath = clt_wenlin.get_wenlin_db_path()
        connection = sqlite3.connect(db_filepath)
        return connection

    def iterate_dictionary_results(self, text, lookup_key):
        connection = self.get_connection()

        language = cloudlanguagetools.languages.Language[lookup_key['language']]
        column_map = {
            cloudlanguagetools.languages.Language.zh_cn: 'simplified',
            cloudlanguagetools.languages.Language.zh_tw: 'traditional',
            cloudlanguagetools.languages.Language.yue: 'traditional',
        }
        column = column_map[language]

        query = f"""SELECT entry FROM words WHERE {column}='{text}'"""
        cur = connection.cursor()
        for row in cur.execute(query):
            entry_json_str = row[0]
            entry_json = json.loads(entry_json_str)
            yield entry_json

        connection.close()

    def collect_definitions(self, generator):
        result = []
        for entry in generator:
            # definitions
            for part_of_speech in entry['parts_of_speech']:
                for definition in part_of_speech['definitions']:
                    result.append(definition['definition'])        
        return result

    def collect_partofspeech(self, generator):
        result = []
        for entry in generator:
            for part_of_speech in entry['parts_of_speech']:
                result.append(part_of_speech['part_of_speech'])

        result = list(set(result))
        result.sort()                
        return result

    def collect_measureword(self, generator):
        result = []
        for entry in generator:
            for part_of_speech in entry['parts_of_speech']:
                for definition in part_of_speech['definitions']:
                    if 'measure_word' in definition:
                        result.append(definition['measure_word'])
        result = list(set(result))
        result.sort()                
        return result                      

    def collect_partofspeech_definitions(self, generator):
        result = {}
        for entry in generator:
            for part_of_speech in entry['parts_of_speech']:
                if part_of_speech['part_of_speech'] not in result:
                    result[part_of_speech['part_of_speech']] = []
                for definition in part_of_speech['definitions']:
                    result[part_of_speech['part_of_speech']].append(definition['definition'])
        return result

    def collect_result_check_empty(self, generator, collect_result_fn, text):
        result = collect_result_fn(generator)

        not_found_error_msg = f'Wenlin: no results found for {text}'

        if type(result) is list:
            if len(result) == 0:
                raise cloudlanguagetools.errors.NotFoundError(not_found_error_msg)
        if type(result) is dict:
            if len(result.keys()) == 0:
                raise cloudlanguagetools.errors.NotFoundError(not_found_error_msg)

        return result

    def get_dictionary_lookup(self, text, lookup_key):
        lookup_type = cloudlanguagetools.constants.DictionaryLookupType[lookup_key['lookup_type']]
        lookup_type_fn_map = {
            cloudlanguagetools.constants.DictionaryLookupType.Definitions: self.collect_definitions,
            cloudlanguagetools.constants.DictionaryLookupType.PartOfSpeech: self.collect_partofspeech,
            cloudlanguagetools.constants.DictionaryLookupType.MeasureWord: self.collect_measureword,
            cloudlanguagetools.constants.DictionaryLookupType.PartOfSpeechDefinitions: self.collect_partofspeech_definitions,
        }

        collect_result_fn = lookup_type_fn_map[lookup_type]
        generator = self.iterate_dictionary_results(text, lookup_key)

        return self.collect_result_check_empty(generator, collect_result_fn, text)


import cloudlanguagetools.service
import cloudlanguagetools.translationlanguage
import requests

import logging
import pprint


logger = logging.getLogger(__name__)


# map from the argos translate codes to our internal language codes
def get_translation_language_enum(language_id):
    argos_language_id_override_map = {
        'zh': 'zh_cn',
        'id': 'id_',
        'pt': 'pt_pt'
    }
    if language_id in argos_language_id_override_map:
        language_id = argos_language_id_override_map[language_id]
    return cloudlanguagetools.languages.Language[language_id]

class LibreTranslateLanguage(cloudlanguagetools.translationlanguage.TranslationLanguage):
    def __init__(self, language_id):
        self.service = cloudlanguagetools.constants.Service.LibreTranslate
        self.service_fee = cloudlanguagetools.constants.ServiceFee.free
        self.language_id = language_id
        self.language = get_translation_language_enum(language_id)

    def get_language_id(self):
        return self.language_id

"""
note: libre translate service offers poor quality translations, disable this service.
"""
class LibreTranslateService(cloudlanguagetools.service.Service):
    BASE_URL = 'http://libretranslate.vocab.ai'

    def __init__(self):
        pass

    def configure(self, config):
        pass

    def get_tts_audio(self, text, voice_key, options):
        raise Exception('not supported')

    def get_tts_voice_list(self):
        return []

    def get_translation(self, text, from_language_key, to_language_key):
        data = {
            'q': text,
            'source': from_language_key,
            'target': to_language_key
        }
        logger.debug(f'translating using parameters: {data}')
        response = requests.post(self.BASE_URL + '/translate', data=data, timeout=cloudlanguagetools.constants.RequestTimeout)
        response_data = response.json()        

        if response.status_code == 200:
            return response_data['translatedText']

        error_message = f'LibreTranslate: could not translate text [{text}] from {from_language_key} to {to_language_key} ({response_data})'
        raise cloudlanguagetools.errors.RequestError(error_message)

    def get_transliteration(self, text, transliteration_key):
        raise Exception('not supported')

    def get_translation_language_list(self):
        result = []

        response = requests.get(self.BASE_URL + '/languages', timeout=cloudlanguagetools.constants.RequestTimeout)
        language_list = response.json()
        for language in language_list:
            try:
                result.append(LibreTranslateLanguage(language['code']))
            except KeyError:
                logger.error(f'could not process translation language for {language}, {language.code}', exc_info=True)

        return result

    def get_transliteration_language_list(self):
        return []



import json
import logging
import tempfile

import cloudlanguagetools.service
import cloudlanguagetools.constants
import cloudlanguagetools.options
import cloudlanguagetools.languages
import cloudlanguagetools.ttsvoice
import cloudlanguagetools.translationlanguage
import cloudlanguagetools.transliterationlanguage
import cloudlanguagetools.dictionarylookup
import cloudlanguagetools.errors


logger = logging.getLogger(__name__)


class TestServiceVoice(cloudlanguagetools.ttsvoice.TtsVoice):
    def __init__(self, audio_language, voice_name, voice_id, service, service_fee):
        self.audio_language = audio_language
        self.gender = cloudlanguagetools.constants.Gender.Female
        self.voice_name = voice_name
        self.voice_id = voice_id
        self.service = service
        self.service_fee = service_fee

    def get_voice_key(self):
        return {
            'voice_id': self.voice_id
        }

    def get_voice_shortname(self):
        return self.voice_name

    def get_options(self):
        return {}

class TestServiceTranslationLanguage(cloudlanguagetools.translationlanguage.TranslationLanguage):
    def __init__(self, language, language_id, service, service_fee):
        self.service = service
        self.service_fee = service_fee
        self.language = language
        self.language_id = language_id

    def get_language_id(self):
        return self.language_id

class TestServiceTransliterationLanguage(cloudlanguagetools.transliterationlanguage.TransliterationLanguage):
    def __init__(self, language, name, transliteration_key, service, service_fee):
        self.service = service
        self.service_fee = service_fee
        self.language = language
        self.transliteration_key = transliteration_key
        self.transliteration_name = name

    def get_transliteration_name(self):
        return self.transliteration_name

    def get_transliteration_shortname(self):
        return self.transliteration_name

    def get_transliteration_key(self):
        return self.transliteration_key

class TestServiceDictionaryLookup(cloudlanguagetools.dictionarylookup.DictionaryLookup):
    def __init__(self, language, name, lookup_key, service, service_fee):
        self.service = service
        self.service_fee = service_fee
        self.language = language
        self.name = name
        self.lookup_key = lookup_key

    def get_lookup_key(self):
        return self.lookup_key

    def get_lookup_name(self):
        return self.name

    def get_lookup_shortname(self):
        return self.name

class TestServiceBase(cloudlanguagetools.service.Service):
    def __init__(self):
        pass

    def configure(self, config):
        pass

    def get_tts_audio(self, text, voice_key, options):
        data_str = json.dumps(
            {
                'type': 'audio',
                'text': text,
                'voice_key': voice_key,
                'options': options
            }
        )

        output_temp_file = tempfile.NamedTemporaryFile()
        output_temp_filename = output_temp_file.name

        f = open(output_temp_filename, 'w')
        f.write(data_str)
        f.close()

        return output_temp_file

    def get_tts_voice_list(self):
        result = []
        result.append(TestServiceVoice(cloudlanguagetools.languages.AudioLanguage.fr_FR, 'Paul', 'paul', 
            self.SERVICE, self.SERVICE_FEE))
        return result

    def get_translation(self, text, from_language_key, to_language_key):
        return json.dumps({
            'text': text,
            'from_language_key': from_language_key,
            'to_language_key': to_language_key
        })
        
    def get_transliteration(self, text, transliteration_key):
        return json.dumps(
            {
                'text': text,
                'transliteration_key': transliteration_key
            }
        )

    def get_translation_language_list(self):
        result = []
        result.append(TestServiceTranslationLanguage(cloudlanguagetools.languages.Language.fr, 'fr', self.SERVICE, self.SERVICE_FEE))
        result.append(TestServiceTranslationLanguage(cloudlanguagetools.languages.Language.en, 'en', self.SERVICE, self.SERVICE_FEE))
        result.append(TestServiceTranslationLanguage(cloudlanguagetools.languages.Language.zh_cn, 'zh', self.SERVICE, self.SERVICE_FEE))
        return result

    def get_transliteration_language_list(self):
        result = []
        result.append(TestServiceTransliterationLanguage(cloudlanguagetools.languages.Language.zh_cn, 'Pinyin', 'pinyin', self.SERVICE, self.SERVICE_FEE))
        return result


    def get_dictionary_lookup_list(self):
        result = []
        result.append(TestServiceDictionaryLookup(cloudlanguagetools.languages.Language.fr, 'French', 'french', self.SERVICE, self.SERVICE_FEE))
        result.append(TestServiceDictionaryLookup(cloudlanguagetools.languages.Language.zh_cn, 'Chinese', 'zh', self.SERVICE, self.SERVICE_FEE))
        return result


    def get_dictionary_lookup(self, text, lookup_key):
        return json.dumps({
            'type': 'dictionary',
            'text': text,
            'lookup_key': lookup_key
        })


class TestServiceA(TestServiceBase):
    SERVICE = cloudlanguagetools.constants.Service.TestServiceA
    SERVICE_FEE = cloudlanguagetools.constants.ServiceFee.free

class TestServiceB(TestServiceBase):
    SERVICE = cloudlanguagetools.constants.Service.TestServiceB
    SERVICE_FEE = cloudlanguagetools.constants.ServiceFee.paid


    


import abc
import json

class TransliterationLanguage(abc.ABC):
    def __init__(self):
        pass

    def get_language_code(self):
        return self.language.name

    def get_language_name(self):
        return self.language.lang_name

    @abc.abstractmethod
    def get_transliteration_name(self):
        pass

    @abc.abstractmethod
    def get_transliteration_shortname(self):
        pass

    @abc.abstractmethod
    def get_transliteration_key(self):
        pass    

    def get_transliteration_id(self):
        return json.dumps({
            'service': self.service.name,
            'key': self.get_transliteration_key()
        })

    def json_obj(self):
        return {
            'service': self.service.name,
            'language_code': self.get_language_code(),
            'language_name': self.get_language_name(),
            'transliteration_name': self.get_transliteration_name(),
            'transliteration_shortname': self.get_transliteration_shortname(),
            'transliteration_key': self.get_transliteration_key(),
            'transliteration_id': self.get_transliteration_id(),
            'service_fee': self.service_fee.name
        } 


import requests
import tempfile
import logging
from typing import List

import cloudlanguagetools.constants
import cloudlanguagetools.ttsvoice

logger = logging.getLogger(__name__)

class Service():
    def __init__(self):
        pass

    def get_service_name(self):
        return self.service.name

    def post_request(self, url, **kwargs):
        kwargs['timeout'] = cloudlanguagetools.constants.RequestTimeout
        return requests.post(url, **kwargs)

    def get_tts_audio_base_post_request(self, url, **kwargs):
        try:
            response = self.post_request(url, **kwargs)
            response.raise_for_status()
            output_temp_file = tempfile.NamedTemporaryFile()
            output_temp_filename = output_temp_file.name            
            with open(output_temp_filename, 'wb') as audio:
                audio.write(response.content)
            return output_temp_file            
        except requests.exceptions.ReadTimeout as exception:
            raise cloudlanguagetools.errors.TimeoutError(f'timeout while retrieving {self.get_service_name()} audio')
        except Exception as exception:
            error_message = f'could not retrieve audio from {self.get_service_name()}'
            logger.exception(error_message)
            raise cloudlanguagetools.errors.RequestError(error_message)

    # used for pre-loading models
    def load_data(self):
        pass

    def get_tts_voice_list(self):
        return []

    def get_tts_voice_list_v3(self) -> List[cloudlanguagetools.ttsvoice.TtsVoice_v3]:
        # the default implementation will convert list of voices to list of TtsVoice_v3
        voices = self.get_tts_voice_list()
        voices_v3 = [cloudlanguagetools.ttsvoice.TtsVoice_v3(
            name=voice.get_voice_shortname(),
            voice_key=voice.get_voice_key(),
            options=voice.get_options(),
            service=voice.service,
            gender=voice.get_gender(),
            audio_languages=[voice.audio_language],
            service_fee=voice.service_fee
        ) for voice in voices]
        return voices_v3

    def get_translation_language_list(self):
        return []

    def get_transliteration_language_list(self):
        return []

    def get_tokenization_options(self):
        return []

    def get_dictionary_lookup_list(self):
        return []


import json
import requests
import tempfile
import logging
import os
import pprint

import cloudlanguagetools.service
import cloudlanguagetools.constants
import cloudlanguagetools.languages
import cloudlanguagetools.translationlanguage
import cloudlanguagetools.transliterationlanguage
import cloudlanguagetools.errors

logger = logging.getLogger(__name__)

class DeepLTranslationLanguage(cloudlanguagetools.translationlanguage.TranslationLanguage):
    def __init__(self, language, language_id):
        self.service = cloudlanguagetools.constants.Service.DeepL
        self.service_fee = cloudlanguagetools.constants.ServiceFee.paid
        self.language = language
        self.language_id = language_id

    def get_language_id(self):
        return self.language_id


class DeepLService(cloudlanguagetools.service.Service):
    def __init__(self):
        self.base_url = 'https://api.deepl.com/v2/translate'

    def configure(self, config):
        self.api_key = config['key']
    
    def get_tts_voice_list(self):
        return []

    def get_headers(self):
        return {
            'Authorization': f'DeepL-Auth-Key {self.api_key}'
        }

    def get_language_enum(self, deepl_language_code):
        lowercase_str = deepl_language_code.lower()
        override_map = {
            'id': 'id_',
            'zh': 'zh_cn',
            'pt': 'pt_pt'
        }
        lowercase_str = override_map.get(lowercase_str, lowercase_str)
        language = cloudlanguagetools.languages.Language
        return language[lowercase_str]


    def get_translation_language_list(self):
        language = cloudlanguagetools.languages.Language
        url = 'https://api.deepl.com/v2/languages'
        response = requests.get(url, headers=self.get_headers(), timeout=cloudlanguagetools.constants.RequestTimeout)
        response.raise_for_status()
        # pprint.pprint(response.json())
        results = []
        for language_entry in response.json():
            try:
                # pprint.pprint(language_entry)
                deepl_language_code = language_entry['language']
                language_enum = self.get_language_enum(deepl_language_code)
                results.append(DeepLTranslationLanguage(language_enum, deepl_language_code))
                # if it's portuguese, replicate the entry for PT-PT and PT-BR
                if language_enum == language.pt_pt:
                    results.append(DeepLTranslationLanguage(language.pt_br, deepl_language_code))
            except Exception as e:
                logger.exception(f'could not process Deepl language entry: {language_entry}')
        return results

    def get_tts_voice_list(self):
        result = []
        return result


    def get_transliteration_language_list(self):
        return []

    def get_translation(self, text, from_language_key, to_language_key):

        override_source_language_map = {
            'PT-PT': 'PT',
            'PT-BR': 'PT'
        }
        from_language_key = override_source_language_map.get(from_language_key, from_language_key)


        params = {
            'auth_key': self.api_key,
            'text': text,
            'source_lang': from_language_key,
            'target_lang': to_language_key
        }
        response = requests.get(self.base_url, params=params, timeout=cloudlanguagetools.constants.RequestTimeout)

        if response.status_code == 200:
            # {'translations': [{'translation': 'Le coût est très bas.'}], 'word_count': 2, 'character_count': 4}
            data = response.json()
            return data['translations'][0]['text']

        error_message = error_message = f'DeepL: could not translate text [{text}] from {from_language_key} to {to_language_key} (status_code: {response.status_code} {response.content})'
        raise cloudlanguagetools.errors.RequestError(error_message)


from openai import OpenAI

import logging
import tempfile
import pydub
import pprint
from typing import List

import cloudlanguagetools.service
import cloudlanguagetools.constants
import cloudlanguagetools.languages
import cloudlanguagetools.options

from cloudlanguagetools.languages import AudioLanguage

logger = logging.getLogger(__name__)

DEFAULT_TTS_SPEED = 1.0

VOICE_OPTIONS = {
            'speed' : {
                'type': cloudlanguagetools.options.ParameterType.number.name,
                'min': 0.25,
                'max': 4.0,
                'default': DEFAULT_TTS_SPEED
            },
            cloudlanguagetools.options.AUDIO_FORMAT_PARAMETER: {
                'type': cloudlanguagetools.options.ParameterType.list.name,
                'values': [
                    cloudlanguagetools.options.AudioFormat.mp3.name,
                    cloudlanguagetools.options.AudioFormat.ogg_opus.name,
                ],
                'default': cloudlanguagetools.options.AudioFormat.mp3.name
            }
}

TTS_SUPPORTED_LANGUAGES = [
            AudioLanguage.af_ZA,
            AudioLanguage.ar_XA,
            AudioLanguage.hy_AM,
            AudioLanguage.az_AZ,
            AudioLanguage.be_BY,
            AudioLanguage.bs_BA,
            AudioLanguage.bg_BG,
            AudioLanguage.ca_ES,
            AudioLanguage.zh_CN,
            AudioLanguage.hr_HR,
            AudioLanguage.cs_CZ,
            AudioLanguage.da_DK,
            AudioLanguage.nl_NL,
            AudioLanguage.en_US,
            AudioLanguage.et_EE,
            AudioLanguage.fi_FI, 
            AudioLanguage.fr_FR, 
            AudioLanguage.gl_ES, 
            AudioLanguage.de_DE, 
            AudioLanguage.el_GR, 
            AudioLanguage.he_IL, 
            AudioLanguage.hi_IN, 
            AudioLanguage.hu_HU, 
            AudioLanguage.is_IS, 
            AudioLanguage.id_ID, 
            AudioLanguage.it_IT, 
            AudioLanguage.ja_JP, 
            AudioLanguage.kn_IN, 
            AudioLanguage.kk_KZ, 
            AudioLanguage.ko_KR, 
            AudioLanguage.lv_LV, 
            AudioLanguage.lt_LT,
            AudioLanguage.mk_MK,
            AudioLanguage.ms_MY,
            AudioLanguage.mr_IN,
            AudioLanguage.ne_NP,
            AudioLanguage.nb_NO,
            AudioLanguage.fa_IR,
            AudioLanguage.pl_PL,
            AudioLanguage.pt_PT,
            AudioLanguage.pt_BR,
            AudioLanguage.ro_RO,
            AudioLanguage.ru_RU,
            AudioLanguage.sr_RS,
            AudioLanguage.sk_SK,
            AudioLanguage.sl_SI,
            AudioLanguage.es_ES,
            AudioLanguage.es_MX,
            AudioLanguage.sw_KE,
            AudioLanguage.sv_SE,
            # AudioLanguage.tl_PH, # need to add tagalog language in AudioLanguages
            AudioLanguage.ta_IN,
            AudioLanguage.ta_LK,
            AudioLanguage.th_TH,
            AudioLanguage.tr_TR,
            AudioLanguage.uk_UA,
            AudioLanguage.ur_PK,
            AudioLanguage.ur_IN,
            AudioLanguage.vi_VN,
            AudioLanguage.cy_GB            
]

class OpenAIVoice(cloudlanguagetools.ttsvoice.TtsVoice):
    def __init__(self, name: str, 
                audio_language: cloudlanguagetools.languages.AudioLanguage,
                gender: cloudlanguagetools.constants.Gender):
        self.name = name
        self.gender = gender
        self.audio_language = audio_language
        self.service = cloudlanguagetools.constants.Service.OpenAI
        self.service_fee = cloudlanguagetools.constants.ServiceFee.paid

    def get_voice_key(self):
        return {
            'name': self.name,
            'language': self.audio_language.name            
        }

    def get_voice_shortname(self):
        return self.name

    def get_options(self):
        return VOICE_OPTIONS

class OpenAIService(cloudlanguagetools.service.Service):
    def __init__(self):
        self.chatbot_model = "gpt-3.5-turbo"

    def configure(self, config):
        self.api_key = config['api_key']
        self.client = OpenAI(api_key=self.api_key)

    def single_prompt(self, prompt, max_tokens):
        messages = [
            {'role': 'user', 'content': prompt}
        ]        

        if max_tokens != None:
            response = self.client.chat.completions.create(model=self.chatbot_model,
            messages=messages,
            max_tokens=max_tokens)    
        else:
            response = self.client.chat.completions.create(model=self.chatbot_model,
            messages=messages)    
        logger.debug(pprint.pformat(response))
        tokens_used = response.usage.total_tokens
        response_text = response.choices[0].message.content
        return response_text, tokens_used

    def full_query(self, messages, max_tokens):
        if max_tokens != None:
            response = self.client.chat.completions.create(model=self.chatbot_model,
            messages=messages,
            max_tokens=max_tokens)    
        else:
            response = self.client.chat.completions.create(model=self.chatbot_model,
            messages=messages)    
        logger.debug(pprint.pformat(response))
        return response

    def speech_to_text(self, filepath, audio_format: cloudlanguagetools.options.AudioFormat):

        if audio_format in [cloudlanguagetools.options.AudioFormat.ogg_opus, cloudlanguagetools.options.AudioFormat.ogg_vorbis]:
            # need to convert to wav first
            sound = pydub.AudioSegment.from_ogg(filepath)
            wav_tempfile = tempfile.NamedTemporaryFile(prefix='cloudlanguagetools_OpenAI_speech_to_text', suffix='.wav')
            sound.export(wav_tempfile.name, format="wav")
            filepath = wav_tempfile.name

        logger.debug(f'opening file {filepath}')
        audio_file= open(filepath, "rb")
        transcript = self.client.audio.transcriptions.create(model="whisper-1", file=audio_file)
        return transcript.text
    
    
    def get_tts_voice_list(self) -> List[OpenAIVoice]:
        result = []

        for audio_language in TTS_SUPPORTED_LANGUAGES:
            result.extend([
                OpenAIVoice('alloy', audio_language, cloudlanguagetools.constants.Gender.Female),
                OpenAIVoice('echo', audio_language, cloudlanguagetools.constants.Gender.Male),
                OpenAIVoice('fable', audio_language, cloudlanguagetools.constants.Gender.Female),
                OpenAIVoice('onyx', audio_language, cloudlanguagetools.constants.Gender.Male),
                OpenAIVoice('nova', audio_language, cloudlanguagetools.constants.Gender.Female),
                OpenAIVoice('shimmer', audio_language, cloudlanguagetools.constants.Gender.Female),
            ])
        return result

    def build_tts_voice_v3(self, voice_name, gender):
        return cloudlanguagetools.ttsvoice.TtsVoice_v3(
            name=voice_name,
            voice_key={
                'name': voice_name
            },
            options=VOICE_OPTIONS,
            service=cloudlanguagetools.constants.Service.OpenAI,
            gender=gender,
            audio_languages=TTS_SUPPORTED_LANGUAGES,
            service_fee=cloudlanguagetools.constants.ServiceFee.paid
        )

    def get_tts_voice_list_v3(self) -> List[cloudlanguagetools.ttsvoice.TtsVoice_v3]:
        # returns list of TtsVoice_v3

        result = [
            self.build_tts_voice_v3('alloy', cloudlanguagetools.constants.Gender.Female),
            self.build_tts_voice_v3('echo', cloudlanguagetools.constants.Gender.Male),
            self.build_tts_voice_v3('fable', cloudlanguagetools.constants.Gender.Female),
            self.build_tts_voice_v3('onyx', cloudlanguagetools.constants.Gender.Male),
            self.build_tts_voice_v3('nova', cloudlanguagetools.constants.Gender.Female),
            self.build_tts_voice_v3('shimmer', cloudlanguagetools.constants.Gender.Female)
        ]
        return result

    def get_tts_audio(self, text, voice_key, options):
        # https://platform.openai.com/docs/guides/text-to-speech
        # https://platform.openai.com/docs/api-reference/audio/createSpeech?lang=python
        
        output_temp_file = tempfile.NamedTemporaryFile()

        speed = options.get('speed', DEFAULT_TTS_SPEED)
        response_format = options.get(cloudlanguagetools.options.AUDIO_FORMAT_PARAMETER, 
            cloudlanguagetools.options.AudioFormat.mp3.name)
        if response_format == cloudlanguagetools.options.AudioFormat.ogg_opus.name:
            response_format = 'opus'

        response = self.client.audio.speech.create(
            model='tts-1-hd',
            voice=voice_key['name'],
            input=text,
            response_format=response_format,
            speed=speed
        )
        response.stream_to_file(output_temp_file.name)

        return output_temp_file



import os
import base64
import tempfile
import logging
import timeit
import cachetools
from typing import List
import cloudlanguagetools.constants
import cloudlanguagetools.languages
import cloudlanguagetools.errors
import cloudlanguagetools.azure
import cloudlanguagetools.google
import cloudlanguagetools.watson
import cloudlanguagetools.naver
import cloudlanguagetools.amazon
import cloudlanguagetools.forvo
import cloudlanguagetools.cereproc
import cloudlanguagetools.vocalware
import cloudlanguagetools.fptai
import cloudlanguagetools.voicen
import cloudlanguagetools.elevenlabs
import cloudlanguagetools.deepl
import cloudlanguagetools.epitran
import cloudlanguagetools.mandarincantonese
import cloudlanguagetools.easypronunciation
import cloudlanguagetools.pythainlp
import cloudlanguagetools.spacy
import cloudlanguagetools.wenlin
import cloudlanguagetools.libretranslate
import cloudlanguagetools.openai
import cloudlanguagetools.encryption
import cloudlanguagetools.translationlanguage

LOAD_TEST_SERVICES_ONLY = os.environ.get('CLOUDLANGUAGETOOLS_CORE_TEST_SERVICES', 'no') == 'yes'

if LOAD_TEST_SERVICES_ONLY:
    import cloudlanguagetools.test_services


class ServiceManager():
    def  __init__(self):
        self.services = {}

        if LOAD_TEST_SERVICES_ONLY:
            self.services[cloudlanguagetools.constants.Service.TestServiceA] = cloudlanguagetools.test_services.TestServiceA()
            self.services[cloudlanguagetools.constants.Service.TestServiceB] = cloudlanguagetools.test_services.TestServiceB()
        else:
            self.services[cloudlanguagetools.constants.Service.Azure] = cloudlanguagetools.azure.AzureService()
            self.services[cloudlanguagetools.constants.Service.Google] = cloudlanguagetools.google.GoogleService()
            self.services[cloudlanguagetools.constants.Service.Watson] = cloudlanguagetools.watson.WatsonService()
            self.services[cloudlanguagetools.constants.Service.Naver] = cloudlanguagetools.naver.NaverService()
            self.services[cloudlanguagetools.constants.Service.Amazon] = cloudlanguagetools.amazon.AmazonService()
            self.services[cloudlanguagetools.constants.Service.Forvo] = cloudlanguagetools.forvo.ForvoService()
            self.services[cloudlanguagetools.constants.Service.CereProc] = cloudlanguagetools.cereproc.CereProcService()
            self.services[cloudlanguagetools.constants.Service.VocalWare] = cloudlanguagetools.vocalware.VocalWareService()
            self.services[cloudlanguagetools.constants.Service.FptAi] = cloudlanguagetools.fptai.FptAiService()
            self.services[cloudlanguagetools.constants.Service.ElevenLabs] = cloudlanguagetools.elevenlabs.ElevenLabsService()
            self.services[cloudlanguagetools.constants.Service.EasyPronunciation] = cloudlanguagetools.easypronunciation.EasyPronunciationService()
            self.services[cloudlanguagetools.constants.Service.Epitran] = cloudlanguagetools.epitran.EpitranService()
            self.services[cloudlanguagetools.constants.Service.DeepL] = cloudlanguagetools.deepl.DeepLService()
            self.services[cloudlanguagetools.constants.Service.PyThaiNLP] = cloudlanguagetools.pythainlp.PyThaiNLPService()
            self.services[cloudlanguagetools.constants.Service.Spacy] = cloudlanguagetools.spacy.SpacyService()
            self.services[cloudlanguagetools.constants.Service.MandarinCantonese] = cloudlanguagetools.mandarincantonese.MandarinCantoneseService()            
            self.services[cloudlanguagetools.constants.Service.Wenlin] = cloudlanguagetools.wenlin.WenlinService()
            self.services[cloudlanguagetools.constants.Service.OpenAI] = cloudlanguagetools.openai.OpenAIService()

    def configure_default(self):
        # use the stored keys to configure services
        self.configure_services(cloudlanguagetools.encryption.decrypt())

    def configure_services(self, config):
        for service_name, value in config.items():
            service_enum = cloudlanguagetools.constants.Service[service_name]
            if service_enum in self.services:
                self.services[service_enum].configure(value)

    def get_language_data_json(self):
        # retrieve all language data (tts, translation, transliteration, etc)
        logging.info('retrieving language data')
        
        logging.info('retrieving language list')
        language_list = self.get_language_list()
        logging.info('retrieving translation list')
        translation_language_list = self.get_translation_language_list()
        logging.info('retrieving transliteration language list')
        transliteration_language_list = self.get_transliteration_language_list()
        logging.info('retrieving tts voice list')
        tts_voice_list = self.get_tts_voice_list()
        logging.info('retrieving tokenization options')
        tokenization_options = self.get_tokenization_options()
        logging.info('retrieving dictionary lookup options')
        dictionary_lookup_options = self.get_dictionary_lookup_options()        

        return {
            'language_list': language_list,
            'translation_options': [option.json_obj() for option in translation_language_list],
            'transliteration_options': [option.json_obj() for option in transliteration_language_list],
            'voice_list': [voice.json_obj() for voice in tts_voice_list],
            'tokenization_options': [option.json_obj() for option in tokenization_options],
            'dictionary_lookup_options': [option.json_obj() for option in dictionary_lookup_options],
        }
        

    def get_language_data_json_v2(self):
        # retrieve all language data (tts, translation, transliteration, etc), sort by free/free+paid
        logging.info('retrieving language data')
        
        logging.info('retrieving translation list')
        translation_language_list = self.get_translation_language_list()
        logging.info('retrieving transliteration language list')
        transliteration_language_list = self.get_transliteration_language_list()
        logging.info('retrieving tts voice list')
        tts_voice_list = self.get_tts_voice_list()
        logging.info('retrieving tokenization options')
        tokenization_options = self.get_tokenization_options()
        logging.info('retrieving dictionary lookup options')
        dictionary_lookup_options = self.get_dictionary_lookup_options()        

        return {
            'premium': {
                'translation_options': [option.json_obj() for option in translation_language_list],
                'transliteration_options': [option.json_obj() for option in transliteration_language_list],
                'voice_list': [voice.json_obj() for voice in tts_voice_list],
                'tokenization_options': [option.json_obj() for option in tokenization_options],
                'dictionary_lookup_options': [option.json_obj() for option in dictionary_lookup_options],
            },
            'free': {
                'translation_options': [option.json_obj() for option in translation_language_list if option.service_fee == cloudlanguagetools.constants.ServiceFee.free],
                'transliteration_options': [option.json_obj() for option in transliteration_language_list if option.service_fee == cloudlanguagetools.constants.ServiceFee.free],
                'voice_list': [voice.json_obj() for voice in tts_voice_list if voice.service_fee == cloudlanguagetools.constants.ServiceFee.free],
                'tokenization_options': [option.json_obj() for option in tokenization_options if option.service_fee == cloudlanguagetools.constants.ServiceFee.free],
                'dictionary_lookup_options': [option.json_obj() for option in dictionary_lookup_options if option.service_fee == cloudlanguagetools.constants.ServiceFee.free],
            }
        }
                

    def get_language_list(self):
        result_dict = {}
        for language in cloudlanguagetools.languages.Language:
            result_dict[language.name] = language.lang_name
        return result_dict

    @cachetools.cached(cache=cachetools.TTLCache(maxsize=1024, ttl=cloudlanguagetools.constants.TTLCacheTimeout))
    def get_tts_voice_list(self):
        result = []
        for key, service in self.services.items():
            logging.info(f'retrieving voice list from {key}')
            result.extend(service.get_tts_voice_list())
        return result

    def get_tts_voice_list_json(self):
        tts_voice_list = self.get_tts_voice_list()
        return [voice.json_obj() for voice in tts_voice_list]

    @cachetools.cached(cache=cachetools.TTLCache(maxsize=1024, ttl=cloudlanguagetools.constants.TTLCacheTimeout))
    def get_tts_voice_list_v3(self):
        result = []
        for key, service in self.services.items():
            logging.info(f'retrieving voice list from {key}')
            result.extend(service.get_tts_voice_list_v3())
        return result

    @cachetools.cached(cache=cachetools.TTLCache(maxsize=1024, ttl=cloudlanguagetools.constants.TTLCacheTimeout))
    def get_translation_language_list(self) -> List[cloudlanguagetools.translationlanguage.TranslationLanguage]:
        result = []
        for key, service in self.services.items():
            result.extend(service.get_translation_language_list())
        return result        

    def get_translation_language_list_json(self):
        """return list of languages supported for translation, using plain objects/strings"""
        language_list = self.get_translation_language_list()
        return [language.json_obj() for language in language_list]

    @cachetools.cached(cache=cachetools.TTLCache(maxsize=1024, ttl=cloudlanguagetools.constants.TTLCacheTimeout))
    def get_transliteration_language_list(self):
        result = []
        for key, service in self.services.items():
            result.extend(service.get_transliteration_language_list())
        return result

    def get_transliteration_language_list_json(self):
        """return list of languages supported for transliteration, using plain objects/strings"""
        language_list = self.get_transliteration_language_list()
        return [language.json_obj() for language in language_list]

    @cachetools.cached(cache=cachetools.TTLCache(maxsize=1024, ttl=cloudlanguagetools.constants.TTLCacheTimeout))
    def get_tokenization_options(self):
        result = []
        for key, service in self.services.items():
            result.extend(service.get_tokenization_options())
        return result

    def get_tokenization_options_json(self):
        """return list of languages supported for tokenization, using plain objects/strings"""
        tokenization_list = self.get_tokenization_options()
        return [tokenization_option.json_obj() for tokenization_option in tokenization_list]

    # dictionary lookups

    @cachetools.cached(cache=cachetools.TTLCache(maxsize=1024, ttl=cloudlanguagetools.constants.TTLCacheTimeout))
    def get_dictionary_lookup_options(self):
        result = []
        for key, service in self.services.items():
            result.extend(service.get_dictionary_lookup_list())
        return result

    def get_dictionary_lookup_options_json(self):
        dictionary_lookup_list = self.get_dictionary_lookup_options()
        return [dict_lookup_option.json_obj() for dict_lookup_option in dictionary_lookup_list]

    def get_tts_audio(self, text, service_name, voice_id, options):
        service_enum = cloudlanguagetools.constants.Service[service_name]
        service = self.services[service_enum]
        return service.get_tts_audio(text, voice_id, options)

    def get_translation(self, text, service_name: str, from_language_key, to_language_key):
        """return text"""
        service_enum = cloudlanguagetools.constants.Service[service_name]
        service = self.services[service_enum]
        return service.get_translation(text, from_language_key, to_language_key)

    def get_all_translations(self, text, from_language, to_language):
        global_starttime = timeit.default_timer()

        translation_language_list = self.get_translation_language_list()

        result = {}
        for service_enum, service in self.services.items():
            service_name = service_enum.name
            # locate from language key
            from_language_entries = [x for x in translation_language_list if x.service.name == service_name and x.get_language_code() == from_language]
            if len(from_language_entries) == 1:
                starttime = timeit.default_timer()
                # this service provides the "from" language in translation list
                from_language_id = from_language_entries[0].get_language_id()
                # locate to language key
                to_language_entries = [x for x in translation_language_list if x.service.name == service_name and x.get_language_code() == to_language]
                if len(to_language_entries) == 1:
                    to_language_id = to_language_entries[0].get_language_id()
                    try:
                        result[service_name] = self.get_translation(text, service_name, from_language_id, to_language_id)
                    except cloudlanguagetools.errors.RequestError:
                        pass # don't do anything
                    except Exception as e:
                        # default exception handler
                        logging.exception(f'could not retrieve translation for service {service_name}, text: {text}')
                    time_diff = timeit.default_timer() - starttime
                    logging.info(f'get_all_translation processing time for {service_name}: {time_diff:.1f}')
        global_time_diff = timeit.default_timer() - global_starttime
        logging.info(f'get_all_translation total processing time: {global_time_diff:.1f}')
        return result

    def get_transliteration(self, text, service_name: str, transliteration_key):
        service_enum = cloudlanguagetools.constants.Service[service_name]
        service = self.services[service_enum]
        return service.get_transliteration(text, transliteration_key)

    def get_tokenization(self, text, service_name: str, tokenization_key):
        service_enum = cloudlanguagetools.constants.Service[service_name]
        service = self.services[service_enum]
        return service.get_tokenization(text, tokenization_key)

    def get_dictionary_lookup(self, text, service_name, lookup_key):
        service_enum = cloudlanguagetools.constants.Service[service_name]
        service = self.services[service_enum]
        return service.get_dictionary_lookup(text, lookup_key)

    def get_breakdown(self, text, tokenization_option, translation_option, transliteration_option):
        
        # first, tokenize
        tokenization_service = tokenization_option['service']
        tokenization_key = tokenization_option['tokenization_key']
        tokenization_result = self.get_tokenization(text, tokenization_service, tokenization_key)

        if translation_option != None:
            translation_service = translation_option['service']
            translation_source_language_id = translation_option['source_language_id']
            translation_target_language_id = translation_option['target_language_id']

        if transliteration_option != None:
            transliteration_service = transliteration_option['service']
            transliteration_key = transliteration_option['transliteration_key']

        # then, enrich tokens with translation and transliteration
        result = []
        for token in tokenization_result:
            entry = {
                'token': token['token'],
                'lemma': token['lemma']
            }

            if token['can_translate'] and translation_option != None:
                # translate this token (lemma)
                entry['translation'] = self.get_translation(token['lemma'], translation_service, translation_source_language_id, translation_target_language_id)

            if token['can_transliterate'] and transliteration_option != None:
                # transliterate the token
                entry['transliteration'] = self.get_transliteration(token['token'], transliteration_service, transliteration_key)

            if 'pos_description' in token:
                entry['pos_description'] = token['pos_description']

            result.append(entry)
        
        return result

    # special handling for pinyin and jyutping
    # ========================================
    
    def get_pinyin_supported_languages(self):
        return [
            cloudlanguagetools.languages.Language.zh_cn,
            cloudlanguagetools.languages.Language.zh_tw,
        ]

    def get_jyutping_supported_languages(self):
        return [
            cloudlanguagetools.languages.Language.yue
        ]        

    def get_pinyin(self, text, tone_numbers, spaces, corrections=[]):
        return self.services[cloudlanguagetools.constants.Service.MandarinCantonese].get_pinyin(text, tone_numbers, spaces, corrections)

    def get_jyutping(self, text, tone_numbers, spaces, corrections=[]):
        return self.services[cloudlanguagetools.constants.Service.MandarinCantonese].get_jyutping(text, tone_numbers, spaces, corrections)

    # LLM APIs
    # ========

    def openai_single_prompt(self, text, max_tokens=None):
        return self.services[cloudlanguagetools.constants.Service.OpenAI].single_prompt(text, max_tokens)

    def openai_full_query(self, messages, max_tokens=None):
        return self.services[cloudlanguagetools.constants.Service.OpenAI].full_query(messages, max_tokens)

    def detect_language(self, text_list):
        """returns an enum from languages.Language"""
        service = self.services[cloudlanguagetools.constants.Service.Azure]
        result = service.detect_language(text_list)
        return result

    def service_cost(self, text, service_name, request_type: cloudlanguagetools.constants.RequestType):
        """return the cost of using a service, in characters"""
        service = cloudlanguagetools.constants.Service[service_name]
        character_length = len(text)
        
        FREE_SERVICES = [
            cloudlanguagetools.constants.Service.MandarinCantonese,
            cloudlanguagetools.constants.Service.LibreTranslate,
            cloudlanguagetools.constants.Service.Epitran,
            cloudlanguagetools.constants.Service.Wenlin,
            cloudlanguagetools.constants.Service.PyThaiNLP,
            cloudlanguagetools.constants.Service.TestServiceA,
        ]

        if service in FREE_SERVICES:
            return 0

        if service == cloudlanguagetools.constants.Service.Naver and request_type == cloudlanguagetools.constants.RequestType.audio:
            return 6 * character_length

        return character_length


import json
import requests
import cloudlanguagetools.constants
import cloudlanguagetools.languages
import pinyin_jyutping


class MandarinCantoneseTransliteration(cloudlanguagetools.transliterationlanguage.TransliterationLanguage):
    def __init__(self, language, conversion_type, tone_numbers, spaces):
        self.service = cloudlanguagetools.constants.Service.MandarinCantonese
        self.service_fee = cloudlanguagetools.constants.ServiceFee.free
        self.language = language
        self.conversion_type = conversion_type
        self.tone_numbers = tone_numbers
        self.spaces = spaces

    def get_transliteration_name(self):
        conversion_type_str = self.conversion_type.capitalize()
        tone_numbers_str = "Diacritics"
        if self.tone_numbers:
            tone_numbers_str = "Tone Numbers"
        spaces_str = ""
        if self.spaces:
            spaces_str = "Spaces"
        return f'{self.language.lang_name} to {conversion_type_str} ({tone_numbers_str} {spaces_str}), {self.service.name}'

    def get_transliteration_shortname(self):
        conversion_type_str = self.conversion_type.capitalize()
        tone_numbers_str = "Diacritics"
        if self.tone_numbers:
            tone_numbers_str = "Tone Numbers"
        spaces_str = ""
        if self.spaces:
            spaces_str = "Spaces"
        return f'{conversion_type_str} ({tone_numbers_str} {spaces_str}), {self.service.name}'        

    def get_transliteration_key(self):
        return {
            'conversion_type': self.conversion_type,
            'tone_numbers': self.tone_numbers,
            'spaces': self.spaces
        }

class MandarinCantoneseService(cloudlanguagetools.service.Service):
    def __init__(self):
        self.pinyin_jyutping = pinyin_jyutping.PinyinJyutping()


    def get_tts_voice_list(self):
        return []

    def get_translation_language_list(self):
        return []

    def get_transliteration_language_list(self):
        result = []
        for tone_numbers in [True, False]:
            for spaces in [True, False]:
                result.append(MandarinCantoneseTransliteration(cloudlanguagetools.languages.Language.zh_cn, 'pinyin', tone_numbers, spaces))
                result.append(MandarinCantoneseTransliteration(cloudlanguagetools.languages.Language.zh_tw, 'pinyin', tone_numbers, spaces))
                result.append(MandarinCantoneseTransliteration(cloudlanguagetools.languages.Language.yue, 'jyutping', tone_numbers, spaces))
        return result

    def get_transliteration(self, text, transliteration_key):
        if transliteration_key['conversion_type'] == 'pinyin':
            return self.pinyin_jyutping.pinyin(text, tone_numbers=transliteration_key['tone_numbers'], spaces=transliteration_key['spaces'])
        elif transliteration_key['conversion_type'] == 'jyutping':
            return self.pinyin_jyutping.jyutping(text, tone_numbers=transliteration_key['tone_numbers'], spaces=transliteration_key['spaces'])

        raise Exception(f"unsupported conversion type: {transliteration_key['conversion_type']}")

    # full access, return all results
    def get_pinyin(self, text, tone_numbers, spaces, corrections):
        if len(corrections) == 0:
            return self.pinyin_jyutping.pinyin_all_solutions(text, tone_numbers, spaces)
        else:
            with_corrections = pinyin_jyutping.PinyinJyutping()
            with_corrections.load_pinyin_corrections(corrections)
            return with_corrections.pinyin_all_solutions(text, tone_numbers, spaces)

    def get_jyutping(self, text, tone_numbers, spaces, corrections):
        if len(corrections) == 0:
            return self.pinyin_jyutping.jyutping_all_solutions(text, tone_numbers, spaces)
        else:
            with_corrections = pinyin_jyutping.PinyinJyutping()
            with_corrections.load_jyutping_corrections(corrections)
            return with_corrections.jyutping_all_solutions(text, tone_numbers, spaces)


import json
import requests
import tempfile
import logging
import os
import base64

import cloudlanguagetools.service
import cloudlanguagetools.constants
import cloudlanguagetools.languages
import cloudlanguagetools.ttsvoice
import cloudlanguagetools.translationlanguage
import cloudlanguagetools.transliterationlanguage
import cloudlanguagetools.errors


def get_audio_language_enum(language_iso, country_iso):
    cereproc_audio_id_map = {
        'ro_ro': 'ro_RO',
        'no_NO': 'nb_NO'
    }
    language_enum_name = f'{language_iso}_{country_iso}'
    if language_enum_name in cereproc_audio_id_map:
        language_enum_name = cereproc_audio_id_map[language_enum_name]
    return cloudlanguagetools.languages.AudioLanguage[language_enum_name]

class CereProcVoice(cloudlanguagetools.ttsvoice.TtsVoice):
    def __init__(self, voice_data):
        self.service = cloudlanguagetools.constants.Service.CereProc
        self.service_fee = cloudlanguagetools.constants.ServiceFee.paid
        self.audio_language = get_audio_language_enum(voice_data['language_iso'], voice_data['country_iso'])
        self.name = voice_data['name']
        self.region = voice_data['region']
        self.accent = voice_data['accent']
        gender_str = voice_data['gender'].capitalize()
        if gender_str == 'Neutral':
            gender_str = 'Male'
        self.gender = cloudlanguagetools.constants.Gender[gender_str]

    def get_voice_shortname(self):
        return f'{self.name} ({self.accent})'

    def get_voice_key(self):
        return {
            'name': self.name
        }

    def get_options(self):
        return {}

class CereProcService(cloudlanguagetools.service.Service):
    def __init__(self):
        self.service = cloudlanguagetools.constants.Service.CereProc

    def configure(self, config):
        self.username = config['username']
        self.password = config['password']
    

    def get_access_token(self):
        combined = f'{self.username}:{self.password}'
        auth_string = base64.b64encode(combined.encode('utf-8')).decode('utf-8')
        headers = {'authorization': f'Basic {auth_string}'}

        auth_url = 'https://api.cerevoice.com/v2/auth'
        response = requests.get(auth_url, headers=headers, 
            timeout=cloudlanguagetools.constants.RequestTimeout)
        response.raise_for_status()

        access_token = response.json()['access_token']        
        return access_token
    
    def get_auth_headers(self):
        headers={'Authorization': f'Bearer {self.get_access_token()}'}
        return headers

    def get_translation_language_list(self):
        return []

    def list_voices(self):
        list_voices_url = 'https://api.cerevoice.com/v2/voices'
        
        response = requests.get(list_voices_url, headers=self.get_auth_headers(), 
            timeout=(cloudlanguagetools.constants.ReadTimeout, cloudlanguagetools.constants.RequestTimeout))
        data = response.json()
        return data['voices']

    def get_tts_voice_list(self):
        result = []

        voice_list = self.list_voices()
        for voice in voice_list:
            try:
                result.append(CereProcVoice(voice))
            except KeyError:
                logging.error(f'could not process voice for {voice}', exc_info=True)

        return result

    def get_tts_audio(self, text, voice_key, options):
        voice_name = voice_key['name']
        url = f'https://api.cerevoice.com/v2/speak?voice={voice_name}&audio_format=mp3'

        ssml_text = f"""<?xml version="1.0" encoding="UTF-8"?>
<speak xmlns="http://www.w3.org/2001/10/synthesis">{text}</speak>""".encode(encoding='utf-8')

        return self.get_tts_audio_base_post_request(url, data=ssml_text, headers=self.get_auth_headers())


    def get_transliteration_language_list(self):
        return []



import logging
import requests
import os

import cloudlanguagetools.service
import cloudlanguagetools.constants
import cloudlanguagetools.languages
import cloudlanguagetools.tokenization

logger = logging.getLogger(__name__)

class SpacyTokenization(cloudlanguagetools.tokenization.Tokenization):
    def __init__(self, language, model_name, variant=None):
        self.language = language
        self.service = cloudlanguagetools.constants.Service.Spacy
        self.service_fee = cloudlanguagetools.constants.ServiceFee.free
        self.model_name = model_name
        self.variant = variant

    def get_tokenization_name(self):
        if self.variant != None:
            result = f'{self.language.lang_name} ({self.variant}), {self.service.name}'
        else:
            result = f'{self.language.lang_name}, {self.service.name}'
        return result

    def get_tokenization_key(self):
        return {
            'model_name': self.model_name
        }

class SpacyService(cloudlanguagetools.service.Service):
    BASE_URL = 'http://spacy-api.vocab.ai'

    def __init__(self):
        self.BASE_URL = os.environ.get('SPACY_URL_OVERRIDE', self.BASE_URL)

    def get_tokenization(self, text, tokenization_key):
        model_name = tokenization_key['model_name']

        response = requests.post(self.BASE_URL + '/v1/tokenize', json={'language': model_name, 'text': text}, timeout=cloudlanguagetools.constants.RequestTimeout)
        response_data = response.json()        

        if response.status_code == 200:
            return response_data

        # raise exception
        raise cloudlanguagetools.errors.RequestError(f'could not generate tokenization model_name {model_name} text: {text}: {response}')


    def get_tokenization_options(self):
        result = [
            SpacyTokenization(cloudlanguagetools.languages.Language.en, 'en'),
            SpacyTokenization(cloudlanguagetools.languages.Language.fr, 'fr'),
            SpacyTokenization(cloudlanguagetools.languages.Language.ja, 'ja'),
            SpacyTokenization(cloudlanguagetools.languages.Language.de, 'de'),
            SpacyTokenization(cloudlanguagetools.languages.Language.es, 'es'),
            SpacyTokenization(cloudlanguagetools.languages.Language.ru, 'ru'),
            SpacyTokenization(cloudlanguagetools.languages.Language.pl, 'pl'),
            SpacyTokenization(cloudlanguagetools.languages.Language.it, 'it'),

        ]
        chinese_language_list = [
            cloudlanguagetools.languages.Language.zh_cn,
            cloudlanguagetools.languages.Language.zh_tw,
            cloudlanguagetools.languages.Language.yue
        ]
        for language in chinese_language_list:
            # chinese variants
            result.extend([
                SpacyTokenization(language, 'zh_jieba', 'Jieba (words)'),
                SpacyTokenization(language, 'zh_char', 'Characters'),
                SpacyTokenization(language, 'zh_pkuseg', 'PKUSeg (words)')
            ])
        return result


import json
import requests
import tempfile
import os
import boto3
import botocore.exceptions
import contextlib

import cloudlanguagetools.service
import cloudlanguagetools.constants
import cloudlanguagetools.options
import cloudlanguagetools.languages
import cloudlanguagetools.ttsvoice
import cloudlanguagetools.translationlanguage
import cloudlanguagetools.transliterationlanguage
import cloudlanguagetools.errors

DEFAULT_VOICE_PITCH = 0
DEFAULT_VOICE_RATE = 100

def get_audio_language_enum(language_code):
    language_map = {
        'arb': 'ar_XA',
        'cmn-CN': 'zh_CN',
        'yue-CN': 'zh_HK'
    }

    language_enum_name = language_code.replace('-', '_')
    if language_code in language_map:
        language_enum_name = language_map[language_code]

    return cloudlanguagetools.languages.AudioLanguage[language_enum_name]

class AmazonVoice(cloudlanguagetools.ttsvoice.TtsVoice):
    def __init__(self, voice_data):
        # print(voice_data)
        # {'Gender': 'Female', 'Id': 'Lotte', 'LanguageCode': 'nl-NL', 'LanguageName': 'Dutch', 'Name': 'Lotte', 'SupportedEngines': ['standard']}
        self.service = cloudlanguagetools.constants.Service.Amazon
        self.service_fee = cloudlanguagetools.constants.ServiceFee.paid
        self.gender = cloudlanguagetools.constants.Gender[voice_data['Gender']]
        self.voice_id = voice_data['Id']
        self.name = voice_data['Name']
        self.audio_language = get_audio_language_enum(voice_data['LanguageCode'])
        self.engine = 'standard'
        if 'neural' in voice_data['SupportedEngines']:
            self.engine = 'neural'

    def get_voice_key(self):
        return {
            'voice_id': self.voice_id,
            'engine': self.engine
        }

    def get_voice_shortname(self):
        return f'{self.name} ({self.engine.capitalize()})'

    def get_options(self):
        return {
            'rate' : {
                'type': 'number',
                'min': 20,
                'max': 200,
                'default': DEFAULT_VOICE_RATE
            },
            'pitch': {
                'type': 'number',
                'min': -50,
                'max': 50,
                'default': DEFAULT_VOICE_PITCH
            },
            cloudlanguagetools.options.AUDIO_FORMAT_PARAMETER: {
                'type': cloudlanguagetools.options.ParameterType.list.name,
                'values': [
                    cloudlanguagetools.options.AudioFormat.mp3.name,
                    cloudlanguagetools.options.AudioFormat.ogg_vorbis.name,
                ],
                'default': cloudlanguagetools.options.AudioFormat.mp3.name
            }            
        }

class AmazonTranslationLanguage(cloudlanguagetools.translationlanguage.TranslationLanguage):
    def __init__(self, language, language_id):
        self.service = cloudlanguagetools.constants.Service.Amazon
        self.service_fee = cloudlanguagetools.constants.ServiceFee.paid
        self.language = language
        self.language_id = language_id

    def get_language_id(self):
        return self.language_id

class AmazonService(cloudlanguagetools.service.Service):
    def __init__(self):
        pass

    def configure(self, config):
        aws_access_key_id = config['AWS_ACCESS_KEY_ID']
        aws_secret_access_key = config['AWS_SECRET_ACCESS_KEY']
        region_name = config['AWS_DEFAULT_REGION']

        self.polly_client = boto3.client("polly", 
            region_name=region_name,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            config=botocore.config.Config(connect_timeout=cloudlanguagetools.constants.RequestTimeout, read_timeout=cloudlanguagetools.constants.RequestTimeout))
        self.translate_client = boto3.client("translate",
            region_name=region_name,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,        
            config=botocore.config.Config(connect_timeout=cloudlanguagetools.constants.RequestTimeout, read_timeout=cloudlanguagetools.constants.RequestTimeout))

    def get_translation(self, text, from_language_key, to_language_key):
        result = self.translate_client.translate_text(Text=text, 
                    SourceLanguageCode=from_language_key, TargetLanguageCode=to_language_key)
        return result.get('TranslatedText')

    def get_tts_audio(self, text, voice_key, options):
        audio_format_str = options.get(cloudlanguagetools.options.AUDIO_FORMAT_PARAMETER, cloudlanguagetools.options.AudioFormat.mp3.name)
        audio_format = cloudlanguagetools.options.AudioFormat[audio_format_str]

        audio_format_map = {
            cloudlanguagetools.options.AudioFormat.mp3: 'mp3',
            cloudlanguagetools.options.AudioFormat.ogg_vorbis: 'ogg_vorbis'
        }

        output_temp_file = tempfile.NamedTemporaryFile()
        output_temp_filename = output_temp_file.name

        pitch = options.get('pitch', DEFAULT_VOICE_PITCH)
        pitch_str = f'{pitch:+.0f}%'
        rate = options.get('rate', DEFAULT_VOICE_RATE)
        rate_str = f'{rate:0.0f}%'

        prosody_tags = f'pitch="{pitch_str}" rate="{rate_str}"'
        if voice_key['engine'] == 'neural':
            # pitch not supported on neural voices
            prosody_tags = f'rate="{rate_str}"'


        ssml_str = f"""<speak>
    <prosody {prosody_tags} >
        {text}
    </prosody>
</speak>"""

        try:
            response = self.polly_client.synthesize_speech(Text=ssml_str, TextType="ssml", OutputFormat=audio_format_map[audio_format], VoiceId=voice_key['voice_id'], Engine=voice_key['engine'])
        except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as error:
            raise cloudlanguagetools.errors.RequestError(str(error))

        if "AudioStream" in response:
            # Note: Closing the stream is important because the service throttles on the
            # number of parallel connections. Here we are using contextlib.closing to
            # ensure the close method of the stream object will be called automatically
            # at the end of the with statement's scope.
            with contextlib.closing(response["AudioStream"]) as stream:
                with open(output_temp_filename, 'wb') as audio:
                    audio.write(stream.read())
                return output_temp_file

        else:
            # The response didn't contain audio data, exit gracefully
            raise cloudlanguagetools.errors.RequestError('no audio stream')


    def get_tts_voice_list(self):
        result = []
        response = self.polly_client.describe_voices()
        # print(response['Voices'])
        for voice in response['Voices']:
            result.append(AmazonVoice(voice))
        return result


    def get_translation_language_list(self):
        result = [
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.af, 'af'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.sq, 'sq'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.am, 'am'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.ar, 'ar'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.hy, 'hy'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.az, 'az'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.bn, 'bn'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.bs, 'bs'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.bg, 'bg'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.ca, 'ca'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.zh_cn, 'zh'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.zh_tw, 'zh-TW'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.hr, 'hr'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.cs, 'cs'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.da, 'da'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.prs, 'fa-AF'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.nl, 'nl'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.en, 'en'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.et, 'et'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.fa, 'fa'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.tl, 'tl'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.fi, 'fi'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.fr, 'fr'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.fr_ca, 'fr-CA'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.ka, 'ka'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.de, 'de'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.el, 'el'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.gu, 'gu'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.ht, 'ht'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.ha, 'ha'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.he, 'he'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.hi, 'hi'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.hu, 'hu'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.is_, 'is'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.id_, 'id'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.it, 'it'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.ja, 'ja'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.kn, 'kn'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.kk, 'kk'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.ko, 'ko'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.lv, 'lv'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.lt, 'lt'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.mk, 'mk'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.ms, 'ms'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.ml, 'ml'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.mt, 'mt'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.mn, 'mn'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.no, 'no'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.fa, 'fa'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.ps, 'ps'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.pl, 'pl'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.pt_pt, 'pt'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.ro, 'ro'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.ru, 'ru'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.sr_cyrl, 'sr'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.si, 'si'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.sk, 'sk'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.sl, 'sl'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.so, 'so'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.es, 'es'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.es_mx, 'es-MX'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.sw, 'sw'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.sv, 'sv'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.tl, 'tl'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.ta, 'ta'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.te, 'te'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.th, 'th'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.tr, 'tr'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.uk, 'uk'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.ur, 'ur'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.uz, 'uz'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.vi, 'vi'),
            AmazonTranslationLanguage(cloudlanguagetools.languages.Language.cy, 'cy')
        ]
        return result

    def get_transliteration_language_list(self):
        result = []
        return result
   


import cloudlanguagetools.service
import cloudlanguagetools.translationlanguage

import logging
# configure argos package location before importing argos
import clt_argostranslate
clt_argostranslate.configure_package_dir()
import argostranslate
import argostranslate.package
import argostranslate.translate

logger = logging.getLogger(__name__)


# map from the argos translate codes to our internal language codes
def get_translation_language_enum(language_id):
    # print(f'language_id: {language_id}')
    argos_language_id_override_map = {
        'zh': 'zh_cn',
        'id': 'id_',
        'pt': 'pt_pt'
    }
    if language_id in argos_language_id_override_map:
        language_id = argos_language_id_override_map[language_id]
    return cloudlanguagetools.languages.Language[language_id]

class ArgosTranslationLanguage(cloudlanguagetools.translationlanguage.TranslationLanguage):
    def __init__(self, language_id):
        self.service = cloudlanguagetools.constants.Service.ArgosTranslate
        self.language_id = language_id
        self.language = get_translation_language_enum(language_id)

    def get_language_id(self):
        return self.language_id

# NOTE: this service is disabled
class ArgosTranslateService(cloudlanguagetools.service.Service):
    def __init__(self):
        pass

    def configure(self, config):
        pass

    def get_tts_audio(self, text, voice_key, options):
        raise Exception('not supported')

    def get_tts_voice_list(self):
        return []

    def get_translation(self, text, from_language_key, to_language_key):

        installed_languages = argostranslate.translate.get_installed_languages()
        from_lang = list(filter(
            lambda x: x.code == from_language_key,
            installed_languages))[0]
        to_lang = list(filter(
            lambda x: x.code == to_language_key,
            installed_languages))[0]
        translation = from_lang.get_translation(to_lang)
        translated_text = translation.translate(text)

        return translated_text

    def get_transliteration(self, text, transliteration_key):
        raise Exception('not supported')

    def get_translation_language_list(self):
        result = []
        installed_languages = argostranslate.translate.get_installed_languages()
        for language in installed_languages:
            try:
                result.append(ArgosTranslationLanguage(language.code))
            except KeyError:
                logger.error(f'could not process translation language for {language}, {language.code}', exc_info=True)

        return result

    def get_transliteration_language_list(self):
        return []



import enum
from strenum import StrEnum

# usage/account related constants 
# ===============================

class RequestType(enum.Enum):
    translation = enum.auto()
    transliteration = enum.auto()
    audio = enum.auto()
    breakdown = enum.auto()
    dictionary = enum.auto()

class UsageScope(enum.Enum):
    def __init__(self, key_str):
        self.key_str = key_str 
    User = ("user")
    Global = ("global")

class UsagePeriod(enum.Enum):
    daily = enum.auto()
    monthly = enum.auto()
    patreon_monthly = enum.auto()
    lifetime = enum.auto()
    recurring = enum.auto()

class ApiKeyType(enum.Enum):
    test = enum.auto()
    patreon = enum.auto()
    trial = enum.auto()
    getcheddar = enum.auto()

class Client(enum.Enum):
    awesometts = enum.auto()
    languagetools = enum.auto()
    hypertts = enum.auto()
    test = enum.auto()

class ServiceFee(enum.Enum):
    free = enum.auto()
    paid = enum.auto()

# what triggered this request (batch / on the fly / editor)
class RequestMode(enum.Enum):
    batch = enum.auto()
    dynamic = enum.auto()
    edit = enum.auto()

class APIVersion(enum.Enum):
    v1 = enum.auto()
    v2 = enum.auto()
    v3 = enum.auto()

# service and language related constants
# ======================================

RequestTimeout = 10 # 10 seconds max
ReadTimeout = 3 # 3 seconds read timeout

TTLCacheTimeout = 86400 # 24 hours

class Service(StrEnum):
    Azure = 'Azure'
    Google = 'Google'
    MandarinCantonese = 'MandarinCantonese'
    EasyPronunciation = 'EasyPronunciation'
    OpenAI = 'OpenAI'
    Watson = 'Watson'
    Naver = 'Naver'
    Amazon = 'Amazon'
    Forvo = 'Forvo'
    CereProc = 'CereProc'
    Epitran = 'Epitran'
    DeepL = 'DeepL'
    VocalWare = 'VocalWare'
    Voicen = 'Voicen'
    FptAi = 'FptAi'
    PyThaiNLP = 'PyThaiNLP'
    Spacy = 'Spacy'
    Wenlin =    'Wenlin'
    LibreTranslate = 'LibreTranslate'
    ElevenLabs = 'ElevenLabs'
    TestServiceA = 'TestServiceA'
    TestServiceB = 'TestServiceB'

class Gender(StrEnum):
    Male = 'Male'
    Female = 'Female'
    Any = 'Any'


class DictionaryLookupType(enum.Enum):
    Definitions = enum.auto()
    PartOfSpeech = enum.auto()
    MeasureWord = enum.auto()
    PartOfSpeechDefinitions = enum.auto()